(* source language ---------------------------------------------------------- *)

type var = string

(* An object language: a simple datatype of 64-bit integer expressions *)
type exp =
  | Var of var            (* string representing an object-language variable *)
  | Const of int64        (* a constant int64 value *)
  | Add of exp * exp      (* sum of two expressions *)
  | Mult of exp * exp     (* product of two expressions *)
  | Neg of exp            (* negation of an expression *)



(* Static context: contains the set of legally scoped variables *)
let context = ["X1"; "X2"; "X3"; "X4"; "X5"; "X6"; "X7"; "X8"]

(* Checks that the source program is well formed *)
let rec static_check (e:exp) : unit =
  begin match e with
    | Var x -> if List.mem x context
               then ()
               else failwith @@ Printf.sprintf "Variable not in scope: %s" x
    | Const _ -> ()
    | Add(e1, e2)
    | Mult(e1, e2) -> static_check e1; static_check e2
    | Neg(e) -> static_check e
  end



(* intermediate representation ---------------------------------------------- *)

(* A stack-oriented intermediate representation *)
type insn =
  | IPushC of int64   (* push an int32 constant onto the stack *)
  | IPushV of var     (* push value of var onto the stack *)
  | IMul              (* multiply the top two values on the stack *)
  | IAdd              (* add the top two values on the stack *)
  | INeg              (* negate the top value on the stack *)


type program = insn list

let rec flatten (e:exp) : program =
  begin match e with
    | Var x -> [IPushV x]
    | Const c -> [IPushC c]
    | Add(e1, e2) -> (flatten e1) @ (flatten e2) @ [IAdd]
    | Mult(e1, e2) -> (flatten e1) @ (flatten e2) @ [IMul]
    | Neg(e) -> (flatten e) @ [INeg]
  end

(* back-end translation to x86 ---------------------------------------------- *)

open X86

(* Demonstrates x86_64 System V AMD64 ABI calling conventions:
   - first six integer/pointer args are passed in:
     rdi, rsi, rdx, rcx, r8, r9
   - arguments seven and higher are pushed onto the stack (right-to-left)
     
   return value in rax

   - callee save: rbp, rbx, r12-r15
   - caller save: all others
*)

let function_prologue : X86.ins list =
  Asm.([ Pushq, [~%Rbp]
       ; Movq, [~%Rsp; ~%Rbp]
       ])

let function_epilogue : X86.ins list =
  Asm.([ Popq, [~%Rbp]
       ; Retq, []
    ])


(* Assumes that x is a well-scoped variable *)
let compile_var (x:var) : X86.operand =
  begin match x with
  | "X1" -> Reg Rdi
  | "X2" -> Reg Rsi
  | "X3" -> Reg Rdx
  | "X4" -> Reg Rcx
  | "X5" -> Reg R08
  | "X6" -> Reg R09
  | "X7" -> Ind3(Lit (Int64.of_int (16)), Rbp)
  | "X8" -> Ind3(Lit (Int64.of_int (24)), Rbp)
  | _ -> failwith @@ Printf.sprintf "unbound variable %s" x
  end


(* compilation strategy 1: use the stack-based ir---------------------------- *)

(* use the x86 stack *)

let compile_op (op:opcode) : X86.ins list =
  Asm.([ Popq, [~%Rax]
       ; Popq, [~%R10]
       ; op, [~%R10; ~%Rax]
       ; Pushq, [~%Rax]
       ])

let compile_neg : X86.ins list =
  Asm.([  Popq, [~%Rax]
       ;  Imulq, [~$(-1); ~%Rax]
       ;  Pushq, [~%Rax]
  ])

let rec compile_insn (i:insn) : X86.ins list =
  begin match i with
    | IPushC c -> Asm.([Pushq, [Imm (Lit c)]])
    | IPushV x -> Asm.([Pushq, [compile_var x]])
    | IMul -> compile_op Imulq
    | IAdd -> compile_op Addq
    | INeg -> compile_neg 
  end
    
let compile1 (src:exp) : X86.prog =
  let ir : program =
    static_check src;
    flatten src
  in
  Asm.[ gtext "_program" @@
      function_prologue
    @ (List.concat (List.map compile_insn ir))
    @ [ Popq, [~%Rax] ]
    @ function_epilogue
  ]
  
(* compilation strategy 2: use a different invariant ------------------------ *)

(* Invariant: compile the expression directly into rax
    - use the stack to remember intermediate results for binary operations
*)

let rec compile_exp (e:exp) : X86.ins list =
  failwith "compile_exp unimplemented"
  
let compile2 (src:exp) : X86.prog =
  failwith "compile2 unimplemented"


